"""
Shifts the dates in a data set according to the date shift map for a patient.
"""

import logging
import os
import sys
import datetime
from pathlib import Path
import re
from typing import Union
# Third-party packages
import pandas as pd
from pandas.errors import OutOfBoundsDatetime
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    makeMap,
                                    map2di,
                                    successiveParents)
# Local packages: Script parameters: General
from drapi.code.drapi.constants.phiVariables import (LIST_OF_PHI_DATES_BO,
                                          LIST_OF_PHI_DATES_NOTES,
                                          LIST_OF_PHI_DATES_OMOP)
# Local packages: Script parameters: General
from common import IRB_NUMBER
# Local packages: Script parameters: Paths
# Local packages: Script parameters: File criteria
from common import BO_PORTION_FILE_CRITERIA

# Arguments
"""
Argument descriptions
`PATIENT_DATE_SHIFT_MAP`:                   A Path object. This is the file path of the date-shift map. It expects a four-columned CSV file.
`LOOK_UP_NAME_MAP_INDEX`:                   An integer. Assuming a typical IDR de-identification map format, `0`` is for the identified, original ID name, and `2` is for the de-identified variable name.
`LOOK_UP_NAME_DE_IDENTIFICATION_FORMAT`:    A string. This is required if `LOOK_UP_NAME_MAP_INDEX` is `2`. Options are "lemur" or "classic"
`LIST_OF_PHI_DATES`:                        A list of strings. This list the column headers (i.e., variable names) that need to have their dates de-identified (i.e., date-shifted).

`USE_MODIFIED_OMOP_DATA_SET`:               A boolean. This indicates which version of the OMOP data set to use, the original or the one that has had its columns converted. These versions are defined by their directory locations in `OMOP_PORTION_DIR_*` and `MODIFIED_OMOP_PORTION_DIR_*`, where `*` is `MAC` or `WIN`.
`MAC_PATHS`:                                A list of Path objects. The Path objects point to the directories that should be processed.
`WIN_PATHS`:                                A list of Path objects. The Path objects point to the directories that should be processed.
`LIST_OF_PORTION_CONDITIONS`:               A list of functions. All these fucntions are evaluated on each of the files in each of the directories defined by `*_PATHS`.
"""
PATIENT_DATE_SHIFT_MAP = Path(r"..\Concatenated Results\data\output\makeDateShiftMap\...\Date Shift Map.CSV")
LOOK_UP_NAME_MAP_INDEX = 2
LOOK_UP_NAME_DE_IDENTIFICATION_FORMAT = "lemur"
LIST_OF_PHI_DATES = LIST_OF_PHI_DATES_BO + LIST_OF_PHI_DATES_NOTES + LIST_OF_PHI_DATES_OMOP

# Arguments: Portion Paths and conditions
MAC_PATHS = [Path(r"..\Concatenated Results\data\output\deIdentify\2023-12-13 13-58-38")]
WIN_PATHS = [Path(r"..\Concatenated Results\data\output\deIdentify\2023-12-13 13-58-38")]

LIST_OF_PORTION_CONDITIONS = [BO_PORTION_FILE_CRITERIA]

# Arguments; General
CHUNK_SIZE = 50000

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `PATIENT_DATE_SHIFT_MAP`: "{PATIENT_DATE_SHIFT_MAP}"
    `MAC_PATHS`: "{MAC_PATHS}"
    `WIN_PATHS`: "{WIN_PATHS}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load de-identification maps for dates, i.e., the date shift maps.
    logger.info("""Loading de-identification maps for dates.""")
    map_ = pd.read_csv(PATIENT_DATE_SHIFT_MAP)
    dateShiftMapDi = map2di(map_, fromColumnIndex=LOOK_UP_NAME_MAP_INDEX, toColumnIndex=3)
    lookUpNameOriginal = map_.columns[0]
    lookUpNameSelected = map_.columns[LOOK_UP_NAME_MAP_INDEX]

    # Create alias reverse look-up dictionary
    from common import VARIABLE_ALIASES
    reverseAliasLookup = {}
    for alias, mainName in VARIABLE_ALIASES.items():
        if mainName in reverseAliasLookup.keys():
            reverseAliasLookup[mainName].append(alias)
        else:
            reverseAliasLookup[mainName] = [alias]

    # Synthesize alias names. If we are looking for the original variable name, these are unchaged, but if we are looking for de-identified variable names, we must syntheisze the names.
    if LOOK_UP_NAME_MAP_INDEX == 0:
        pass
    elif LOOK_UP_NAME_MAP_INDEX == 2:
        listOfLookUpNameAliases = []
        for alias in reverseAliasLookup[lookUpNameOriginal]:
            df = makeMap(IDset={},
                         IDName="PatientKey",
                         startFrom=1,
                         irbNumber=IRB_NUMBER,
                         suffix="",
                         columnSuffix="",
                         logger=logger)
            lookUpNameAlias = df.columns[-1]
            listOfLookUpNameAliases.append(lookUpNameAlias)
    else:
        msg = """Invalid value for `LOOK_UP_NAME_MAP_INDEX`. See script arguments."""
        logger.error(msg)
        raise Exception(msg)

    def string2datetime(string: str) -> Union[datetime.datetime, pd.Timestamp]:
        """
        Attempts to convert a string to a pandas datetime object. If it's out of range it converts it to a Python datetime object.

        The Python datetime conversion works as follows:
        Using regex it detects if the string is one of the following formats:
            - Date and time with seconds-precision
            - Date and time with minutes-precision
            - Date
        NOTE: Additional formats can be implemented by adding the corresponding regular expression and `strptime` call combination.
        """
        try:
            newValue = pd.to_datetime(string)
        except OutOfBoundsDatetime as err:
            _ = err
            pattern1 = r"(?P<newValue>\d+[\/-]\d+[\/-]\d+\W+\d+:\d+:\d+)"   # Second-precision
            pattern2 = r"(?P<newValue>\d+[\/-]\d+[\/-]\d+\W+\d+:\d+)"  # Minutes-precision
            pattern3 = r"(?P<newValue>\d+[\/-]\d+[\/-]\d)"  # Days-precision
            reObj1 = re.match(pattern1, string)
            reObj2 = re.match(pattern2, string)
            reObj3 = re.match(pattern3, string)
            if reObj1:
                newValue = datetime.datetime.strptime(string, "%m/%d/%Y %H:%M:%S")
            elif reObj2:
                newValue = datetime.datetime.strptime(string, "%m/%d/%Y %H:%M")
            elif reObj3:
                newValue = datetime.datetime.strptime(string, "%m/%d/%Y")
        return newValue

    def datetimeOffset(timeObj: Union[datetime.datetime, pd.Timestamp], offset: int):
        """
        `offset`, an int, is the number of days.
        """
        if (pd.Timestamp.min < timeObj < pd.Timestamp.max):
            newValue = timeObj + pd.DateOffset(offset)
        else:
            newValue = timeObj + datetime.timedelta(days=offset)
        return newValue

    logger.info("""De-identifying files.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                logger.info("""  ..  Reading file to count the number of chunks.""")
                numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                logger.info(f"""  ..  This file has {numChunks} chunks.""")
                dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logger.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in LIST_OF_PHI_DATES:
                            logger.info("""  ..  ..  Column must be de-identified. Date-shifting values.""")
                            if lookUpNameSelected in dfChunk.columns:
                                lookUpName = lookUpNameSelected
                            elif any([alias in dfChunk.columns for alias in listOfLookUpNameAliases]):
                                for alias in listOfLookUpNameAliases:
                                    if alias in dfChunk.columns:
                                        break
                                lookUpName = alias
                            else:
                                msg = f"""The variable used to look up the date shift offset was not found. We searched for "{lookUpNameSelected}" and any of its aliases: {'", "'.join(listOfLookUpNameAliases)}"""
                                logger.error(msg)
                                raise Exception(msg)
                            dateShiftValues = dfChunk[lookUpName].apply(lambda lookUpID: dateShiftMapDi[lookUpID])
                            column = dfChunk[columnName].apply(string2datetime)
                            df = pd.concat([column, dateShiftValues], axis=1)
                            series = df.apply(lambda el: (el[columnName], el[lookUpName]), axis=1)
                            dfChunk[columnName] = series.apply(lambda tu: datetimeOffset(timeObj=tu[0], offset=tu[1]))
                            dfChunk = dfChunk.rename(columns={columnName: f"De-identified {columnName}"})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""    This file does not need to be processed.""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
HIPPA disclosure template  # TODO
"""

import datetime as dt
import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.drapi import getTimestamp, successiveParents, makeDirPath

# Arguments: Script-specific
# Enter full path to the file with MRNs
FILE_PATH_STRING = "data/output"
file_path = Path(FILE_PATH_STRING)
# Enter file name. The file should be within `file_path`
FILE_NAME_STRING = "MRNs.csv"
file = Path(FILE_NAME_STRING)
# Enter the column name that contains MRNs
column = 'MRN_JAX'
# Enter GNV or JAX to denote whether MRNs are GNV or JAX
site = 'JAX'
# Enter IRB number
irb = 'IRB202202467'
# Enter the PI name
PI_name = 'Jiang Bian'
# Enter PI email
PI_email = 'bianjiang@ufl.edu'
# Analyst
analyst = 'Herman Autore'
# Analyst initials
ANALYST_INITIALS = "HA"
# Analyst email
analyst_email = 'haut0001@shands.ufl.edu'
# Analyst username
analyst_username = 'haut0001'
# Output file prefix
out = f"{ANALYST_INITIALS}_{irb}_{site}_"
# HIPPA directory
hipaa_dir = Path(r'\\shandsdfs.shands.ufl.edu\FILES\SHARE\DSS\IDR Data Requests\HIPAA_Disclosure_files')

# Arguments: General
PROJECT_DIR_DEPTH = 2
IRB_DIR_DEPTH = 2
IDR_DATA_REQUEST_DIR_DEPTH = 5

LOG_LEVEL = "DEBUG"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)

# Directory creation: General
makeDirPath(runLogsDir)


def left_padding(str1):
    """
    Pads a numeric string with leading 0s up to a length of eight (8).
    """
    while (len(str1) < 8):
        str1 = '0' + str1
    return str1


def hipaa_disclosure():
    today = dt.date.today()
    hipaa_disclosure_date_field = today.strftime("%m%d%Y")
    if (site == 'GNV'):
        hipaa_disclosure_header = ['H', analyst_username, analyst_email, irb, '102', 'RE', analyst, hipaa_disclosure_date_field, 'ALL', 'N', ' ', PI_name, '', '', '', '', '', '', '', '', PI_email]
    elif (site == 'JAX'):
        hipaa_disclosure_header = ['H', analyst_username, analyst_email, irb, '998', 'RE', analyst, hipaa_disclosure_date_field, 'ALL', 'N', ' ', PI_name, '', '', '', '', '', '', '', '', PI_email]
    fpath = os.path.join(file_path, file)
    logging.debug(fpath)
    mrns = pd.read_csv(fpath)
    mrns = mrns[[column]].drop_duplicates()
    mrns[column] = mrns[column].fillna(0)
    mrns = mrns[mrns[column] != 0]
    mrns[column] = mrns[column].astype(int)
    mrns[column] = mrns[column].astype(str)
    if (site == 'JAX'):
        mrns[column] = mrns.apply(lambda row: left_padding(row[column]), axis=1)
    mrns = mrns[column].tolist()
    hipaa_disclosure_df = pd.DataFrame(columns=hipaa_disclosure_header)
    hipaa_disclosure_df[hipaa_disclosure_df.columns[1]] = mrns
    hipaa_disclosure_df[hipaa_disclosure_df.columns[0]] = "M"
    hipaa_file = out + str(today) + '.csv'
    hipaa_disclosure_df.to_csv(os.path.join(file_path, hipaa_file), index=False)
    return


if __name__ == '__main__':
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `projectDir`: "{projectDir}".""")

    # Script
    hipaa_disclosure()

    # End script
    logging.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
Concatenates files to create i2b2 `PATIENT_NUM` and `ENCOUNTER_NUM` maps.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
ENCOUNTER_MAP_DIR_PATH = Path(r"..\Intermediate Results\BO Data Portion\data\output\getData\2023-11-27 16-30-11")
PATIENT_MAP_DIR_PATH = Path(r"..\Intermediate Results\BO Data Portion\data\output\getData\2023-11-27 16-45-28")

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"
connection = create_engine(conStr).connect().execution_options(stream_results=True)

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    `ENCOUNTER_MAP_DIR_PATH`: "{ENCOUNTER_MAP_DIR_PATH}"
    `PATIENT_MAP_DIR_PATH`: "{PATIENT_MAP_DIR_PATH}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load encounter map
    encounterMap = pd.DataFrame()
    for fpath in ENCOUNTER_MAP_DIR_PATH.iterdir():
        if fpath.suffix.lower() == ".csv":
            df = pd.read_csv(fpath)
            encounterMap = pd.concat([encounterMap, df])
        else:
            pass

    # Load patient map
    patientMap = pd.DataFrame()
    for fpath in PATIENT_MAP_DIR_PATH.iterdir():
        if fpath.suffix.lower() == ".csv":
            df = pd.read_csv(fpath)
            patientMap = pd.concat([patientMap, df])
        else:
            pass

    # QA
    n1 = len(encounterMap)
    n2 = len(patientMap)
    nu1 = len(encounterMap.iloc[:, 0].unique())
    nu2 = len(encounterMap.iloc[:, 1].unique())
    nu3 = len(patientMap.iloc[:, 0].unique())
    nu4 = len(patientMap.iloc[:, 1].unique())

    logger.info(f"""QA statistics:
encounter map length:           {n1:,}
num unique i2b2 encounter IDs:  {nu1:,}
num unique other encounter IDs: {nu2:,}
patient map length:             {n2:,}
num unique i2b2 patient IDs:    {nu3:,}
num unique other patient IDs:   {nu4:,}.""")

    # Save maps
    encounterpath = runOutputDir.joinpath("i2b2 Encounter Map.CSV")
    patientpath = runOutputDir.joinpath("i2b2 Patient Map.CSV")

    encounterMap.to_csv(encounterpath, index=False)
    patientMap.to_csv(patientpath, index=False)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Makes de-identification maps from scratch.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# TODO Needs to combine similar IDs, like different providers IDs.
"""

import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    makeMap,
                                    successiveParents)
# Local packages: Script parameters: General
from common import (IRB_NUMBER,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    VARIABLE_SUFFIXES)
# Project parameters: Portion paths
from common import (BO_PORTION_DIR_MAC, BO_PORTION_DIR_WIN,
                    I2B2_PORTION_DIR_MAC, I2B2_PORTION_DIR_WIN,
                    NOTES_PORTION_DIR_MAC, NOTES_PORTION_DIR_WIN,
                    OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN)
# Project parameters: File criteria
from common import (BO_PORTION_FILE_CRITERIA,
                    I2B2_PORTION_FILE_CRITERIA,
                    NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA)

# Arguments
LOG_LEVEL = "DEBUG"

MAC_PATHS = [BO_PORTION_DIR_MAC,
             I2B2_PORTION_DIR_MAC,
             NOTES_PORTION_DIR_MAC,
             OMOP_PORTION_DIR_MAC]
WIN_PATHS = [BO_PORTION_DIR_WIN,
             I2B2_PORTION_DIR_WIN,
             NOTES_PORTION_DIR_WIN,
             OMOP_PORTION_DIR_WIN]

LIST_OF_PORTION_CONDITIONS = [BO_PORTION_FILE_CRITERIA,
                              I2B2_PORTION_FILE_CRITERIA,
                              NOTES_PORTION_FILE_CRITERIA,
                              OMOP_PORTION_FILE_CRITERIA]

SETS_PATH = Path(r"..\Concatenated Results\data\output\getIDValues\...\Set Files")

CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Get set of values
    # Imported from "getIDValues.py"

    # Map values
    for file in SETS_PATH.iterdir():
        variableName = file.stem
        logging.info(f"""  Working on variable "{variableName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Read file
        try:
            df = pd.read_table(file, header=None)
            series = df[0]
        except EmptyDataError as err:
            _ = err
            series = pd.Series()
        values = set(series.values)
        # Map contents
        deIdIDSuffix = VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"]
        map_ = makeMap(IDset=values, IDName=variableName, startFrom=1, irbNumber=IRB_NUMBER, suffix=deIdIDSuffix, columnSuffix=variableName, logger=logging.getLogger())
        # Save map
        mapPath = runOutputDir.joinpath(f"{variableName} map.csv")
        map_.to_csv(mapPath, index=False)
        logging.info(f"""    De-identification map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Utility functions for processing the OMOP Data Pull configuration file.
"""

import os
import sys
from pathlib import Path
# Third-party packages
import yaml
# Custom packages
from drapi.code.drapi.drapi import successiveParents


def interpretPath(pathAsString: str) -> str:
    """
    Makes sure path separators are appropriate for the current operating system.
    """
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        newPathAsString = pathAsString.replace("\\", "/")
    elif operatingSystem == "linux":
        newPathAsString = pathAsString.replace("\\", "/")
    elif operatingSystem == "win32":
        newPathAsString = pathAsString.replace("/", "\\")
    else:
        raise Exception("Unsupported operating system")
    return newPathAsString


def editConfig(inputPath: Path, outputPath: Path, timestamp: str) -> None:
    """
    Edits a YAML config file so that the output paths have a timestamp according to the following formats:
    Category 1 Directory: "parent1/parenti/parentn-1/parentn" --> "parent1/parenti/parentn-1/parentn/<timestamp>/"
    Category 2 Directory: "parent1/parenti/parentn-1/parentn" --> "parent1/parenti/parentn-1/<timestamp>/parentn/"

    Category 1 directories:
      - mapping_location
    Category 2 directories:
      - deidentified_file_location
      - identified_file_location
    """
    with open(inputPath) as file:
        configFile = yaml.safe_load(file)

    # Get paths as strings
    identified_file_location_str = configFile["data_output"]["identified_file_location"]
    deidentified_file_location_str = configFile["data_output"]["deidentified_file_location"]
    mapping_location_str = configFile["data_output"]["mapping_location"]

    # Make sure path separators are OS-appropriate
    identified_file_location_str2 = interpretPath(identified_file_location_str)
    deidentified_file_location_str2 = interpretPath(deidentified_file_location_str)
    mapping_location_str2 = interpretPath(mapping_location_str)

    # Add timestamp to path as subfolder
    pathDict = {"identified_file_location": identified_file_location_str2,
                "deidentified_file_location": deidentified_file_location_str2,
                "mapping_location": mapping_location_str2}
    CATEGORY_1 = ["mapping_location"]
    CATEGORY_2 = ["identified_file_location", "deidentified_file_location"]
    sep = os.sep
    for pathName, pathStr in pathDict.items():
        pathObj = Path(pathStr).absolute()
        if pathName in CATEGORY_1:
            newPath = pathObj.joinpath(timestamp)
        elif pathName in CATEGORY_2:
            pathDirName = pathObj.name
            pathParent, _ = successiveParents(pathObj, 1)
            newPath = pathParent.joinpath(timestamp).joinpath(pathDirName)
        else:
            raise Exception(f"""Unexpected value: "{pathName}".""")
        configFile["data_output"][pathName] = str(newPath) + sep

    with open(outputPath, "w") as file:
        yaml.dump(configFile, file)



"""
Functions used in the mapping or de-identification of values, namely "convertColumnsHash.py" in the "Concatenated Results" template/module.
"""

import random
import string
from typing_extensions import (Callable,
                               Collection,
                               Tuple,
                               Union)
# Third-party packages
import numpy as np
import pandas as pd


def encryptValue1(value: Union[float, int, str],
                  secret: Union[int, Collection[int]]):
    """
    Additive encryption.
    Example:
    ```
    encryptValue1(value='123456789', secret=1)
    # 123456790
    ```
    """
    intsAndFloats = (int,
                     np.integer,
                     float,
                     np.floating)
    if isinstance(value, str):
        newValue = float(value) + secret
    elif isinstance(secret, intsAndFloats):
        newValue = value + secret
    else:
        newValue = value + secret
    return newValue


def encryptValue2(value: Union[int, str],
                  secret: str):
    r"""
    Encrypt with character-wise XOR operation of both operands, with the second operand rotating over the set of character-wise values in `secretkey`.
    Example:
    ```
    encryptValue1(value='123456789', secret='password')
    # 'AS@GBYE\I'
    ```
    """
    if isinstance(value, (int, str)):
        valueInOrd = []
        for el in str(value):
            valueInOrd.append(ord(el))
        resultList = []
        for it, integer in enumerate(valueInOrd):
            result = integer ^ ord(secret[it % len(secret)])
            resultList.append(result)
        newValue = "".join([chr(el) for el in resultList])
    else:
        raise Exception("`value` must be of type `int` or `str`.")
    return newValue


def encryptValue3(value: Union[int, str],
                  secret: int):
    """
    Encrypt with whole-value XOR operation. Requires both operands to be integers.
    Example:
    ```
    encryptValue1(value=123456789, secret=111111111)
    # 1326016938
    ```
    """
    if isinstance(value, (int,)):
        newValue = value ^ secret
    else:
        raise Exception(f"""`value` must be of type `int`. Received type "{type(value)}".""")
    return newValue


def deIdentificationFunction(encryptionFunction,
                             irbNumber,
                             suffix,
                             value):
    """
    This function creates a label from an encryption value.
    """
    if pd.isna(value):
        deIdentifiedValue = ""
    else:
        newValue = encryptionFunction(value)
        deIdentifiedValue = f"{irbNumber}_{suffix}_{newValue}"
    return deIdentifiedValue


def functionFromSettings(ENCRYPTION_TYPE: int,
                         ENCRYPTION_SECRET: Union[int, str],
                         IRB_NUMBER: str,
                         suffix) -> Tuple[Union[int, str], Callable]:
    """
    """
    ENCRYPTION_TYPE = int(ENCRYPTION_TYPE)
    ENCRYPTION_SECRET = str(ENCRYPTION_SECRET)
    # Set Parameter: Encryption function: Core
    if ENCRYPTION_TYPE == 1:
        encryptionFunction0 = encryptValue1
    elif ENCRYPTION_TYPE == 2:
        encryptionFunction0 = encryptValue2
    elif ENCRYPTION_TYPE == 3:
        encryptionFunction0 = encryptValue3
    else:
        raise Exception(f"""Argument `ENCRYPTION_TYPE` must be one of {{1, 2, 3}}, instead got "{ENCRYPTION_TYPE}".""")
    # Set Parameter: Encryption secret
    if ENCRYPTION_SECRET.lower() == "random":
        if ENCRYPTION_TYPE in [1, 3]:
            # Integer
            encryptionSecret = np.random.randint(1000000, 10000000)
        elif ENCRYPTION_TYPE in [2]:
            # Alphanumeric
            encryptionSecret = ''.join(random.choice(string.ascii_letters + string.digits + string.punctuation + string.whitespace) for _ in range(15))
        else:
            raise Exception(f"""Argument `ENCRYPTION_TYPE` must be one of {{1, 2, 3}}, instead got "{ENCRYPTION_TYPE}".""")
    elif ENCRYPTION_SECRET.isnumeric():
        encryptionSecret = int(ENCRYPTION_SECRET)
    else:
        encryptionSecret = ENCRYPTION_SECRET

    # Set parameter: Encryption function: Wrapper
    def variableFunction(value: Union[int, str]):
        """
        This is a wrapper function for `deIdentificationFunction`
        """
        # Pre-process values
        if pd.isna(value):
            deIdentifiedValue = ""
        else:
            if value < 0:
                newValue = 0
            else:
                if isinstance(value, float):
                    value1 = int(value)
                else:
                    value1 = value
                newValue = encryptionFunction0(value1, encryptionSecret)

            deIdentifiedValue = f"{IRB_NUMBER}_{suffix}_{newValue}"

        return deIdentifiedValue

    return (encryptionSecret, variableFunction)



"""
Placeholder file for reference. To perform this workflow see the script located at

"IDR Data Requests/ACTIVE RDRs/Bian/IRB202202467/Data Request 04 - 2024-04-04/Intermediate Results/BO Portion Template/code/mapIDsOFtoMRN.py"

"""

print("Not implemented! Please see this module's docstring.")



"""
Get the set of ID values for all variables to de-identify.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""

__all__ = ["runIntermediateDataDir"]

import csv
import json
import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    sortIntegersAndStrings,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import VARIABLE_NAME_TO_FILE_NAME_DICT
# Project parameters: General
from common import (COLUMNS_TO_DE_IDENTIFY,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    DATA_TYPES_DICT)
# Project parameters: Portion paths and criteria
from common import (MODIFIED_OMOP_PORTION_DIR_MAC, MODIFIED_OMOP_PORTION_DIR_WIN,
                    NOTES_PORTION_DIR_MAC, NOTES_PORTION_DIR_WIN,
                    OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN)
# Project parameters: Criteria
from common import (NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA)


# Arguments
SETS_INTERMEDIATE_PATH = None

CHUNK_SIZE = 50000

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Arguments: Portion Paths and conditions
if USE_MODIFIED_OMOP_DATA_SET:
    OMOPPortionDirMac = MODIFIED_OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = MODIFIED_OMOP_PORTION_DIR_WIN
else:
    OMOPPortionDirMac = OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = OMOP_PORTION_DIR_WIN

PORTION_PATHS_MAC = {"Notes": NOTES_PORTION_DIR_MAC,
                     "OMOP": OMOPPortionDirMac}
PORTION_PATHS_WIN = {"Notes": NOTES_PORTION_DIR_WIN,
                     "OMOP": OMOPPortionDirWin}

DICT_OF_PORTION_CONDITIONS = {"Notes": NOTES_PORTION_FILE_CRITERIA,
                              "OMOP": OMOP_PORTION_FILE_CRITERIA}

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in PORTION_PATHS_MAC.values()]) or all([path.exists() for path in PORTION_PATHS_WIN.values()])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        dictOfPortionPaths = PORTION_PATHS_MAC.copy()
    elif operatingSystem == "win32":
        dictOfPortionPaths = PORTION_PATHS_WIN.copy()
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Match portion paths and conditions
    portionPathsAndConditions = {portionName: (dictOfPortionPaths[portionName], DICT_OF_PORTION_CONDITIONS[portionName]) for portionName in dictOfPortionPaths.keys()}
    check1 = [pn1 in DICT_OF_PORTION_CONDITIONS.keys() for pn1 in dictOfPortionPaths.keys()]
    check2 = [pn2 in dictOfPortionPaths.keys() for pn2 in DICT_OF_PORTION_CONDITIONS.keys()]
    assert sum(check1) == len(check1), "Not all portion paths are associated with a portion condition"
    assert sum(check2) == len(check2), "Not all portion conditions are associated with a portion path"

    # Misc
    columnSetsVarsDiFname = "Column Sets Dict.JSON"

    # Get set of values
    if SETS_INTERMEDIATE_PATH:
        logging.info(f"""Using the set of values previously collected from "{SETS_INTERMEDIATE_PATH}".""")
        with open(SETS_INTERMEDIATE_PATH.joinpath(columnSetsVarsDiFname)) as file:
            columnSetsVarsDi = json.loads(file.read())
    else:
        logging.info("""Getting the set of values for each variable to de-identify.""")
        columnSetsVarsDi = {columnName: {"fpath": runIntermediateDataDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[columnName]}.txt"),
                                         "fileMode": "w",
                                         "portionName": None,
                                         "collected": False} for columnName in COLUMNS_TO_DE_IDENTIFY}
        for portionName, (directory, fileConditions) in portionPathsAndConditions.items():
            # Act on directory
            logging.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
            for file in directory.iterdir():
                logging.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
                conditions = [condition(file) for condition in fileConditions]
                if all(conditions):
                    # Read file
                    logging.info("""    File has met all conditions for processing.""")
                    numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE)])
                    dfChunks = readDataFile(file, chunksize=CHUNK_SIZE)
                    for it, dfChunk in enumerate(dfChunks, start=1):
                        dfChunk = pd.DataFrame(dfChunk)
                        logging.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                        for columnName in dfChunk.columns:
                            logging.info(f"""  ..    Working on column "{columnName}".""")
                            if columnName in COLUMNS_TO_DE_IDENTIFY:
                                logging.info("""  ..  ..  Column must be de-identified. Collecting values.""")
                                dataType = DATA_TYPES_DICT[columnName]
                                series = dfChunk[columnName]
                                series = series.dropna()
                                series = series.drop_duplicates()
                                if dataType == "Datetime":
                                    values = series.sort_values()
                                    quoting = csv.QUOTE_NONE
                                elif dataType == "Numeric":
                                    series = series.astype(float).astype("Int64")
                                    values = series.sort_values()
                                    quoting = csv.QUOTE_NONE
                                elif dataType == "String":
                                    values = series.astype(str).sort_values()
                                    quoting = csv.QUOTE_ALL
                                elif dataType == "Numeric_Or_String":
                                    mask = series.apply(lambda el: isinstance(el, float))
                                    series.loc[mask[mask].index] = series[mask].astype(int)
                                    values = sortIntegersAndStrings(series.to_list())
                                    values = pd.Series(values)
                                    quoting = csv.QUOTE_ALL
                                else:
                                    raise Exception(f"""Unexpected `dataType` value: "{dataType}".""")
                                columnSetFpath = columnSetsVarsDi[columnName]["fpath"]
                                columnSetFileMode = columnSetsVarsDi[columnName]["fileMode"]
                                # logging.info(f"""  ..  ..  Values table preview:\n{values.head()}.""")
                                values.to_csv(path_or_buf=columnSetFpath,
                                              quoting=quoting,
                                              index=False,
                                              header=False,
                                              mode=columnSetFileMode)
                                # logging.info(f"""  ..  ..  Preview of table after saving:\n{pd.read_table(columnSetFpath)}.""")
                                columnSetsVarsDi[columnName]["fileMode"] = "a"
                                columnSetsVarsDi[columnName]["portionName"] = portionName
                                columnSetsVarsDi[columnName]["collected"] = True
                                logging.info(f"""  ..  ..  Values saved to "{columnSetFpath.absolute().relative_to(rootDirectory)}" in the project directory.""")
                else:
                    logging.info("""    This file does not need to be processed.""")

        columnSetsVarsDiPath = runIntermediateDataDir.joinpath("Metadata", columnSetsVarsDiFname)
        makeDirPath(columnSetsVarsDiPath.parent)
        columnSetsVarsDiSerializable = columnSetsVarsDi.copy()
        for columnName, di in columnSetsVarsDiSerializable.items():
            columnSetsVarsDiSerializable[columnName]["fpath"] = str(di["fpath"])
        with open(columnSetsVarsDiPath, "w") as file:
            file.write(json.dumps(columnSetsVarsDi))

    # Drop variables that weren't found
    for columnName in COLUMNS_TO_DE_IDENTIFY:
        if columnName in columnSetsVarsDi.keys():
            if columnSetsVarsDi[columnName]["collected"] is False:
                columnSetsVarsDi.pop(columnName)

    # Remove duplicates from set files and save according to data type
    logging.info("Removing duplicates from set files and saving according to data type.")
    columnSetsVarsLi = sorted(list(columnSetsVarsDi.items()), key=lambda tu: tu[0].lower())
    for columnName, fileDi in columnSetsVarsLi:
        logging.info(f"""  Working on variable "{columnName}".""")
        fpath = fileDi["fpath"]
        try:
            series = pd.read_table(fpath, header=None)[0]
        except EmptyDataError as err:
            _ = err
            series = pd.Series(dtype=str)
        # Save according to data type
        portionName = fileDi["portionName"]
        dataType = DATA_TYPES_DICT[columnName]
        if dataType == "Datetime":
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "Numeric":
            series = series.astype(float).astype("Int64")
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "String":
            quoting = csv.QUOTE_ALL
        elif dataType == "Numeric_Or_String":
            quoting = csv.QUOTE_ALL
        else:
            raise Exception(f"""Unexpected `dataType` value: "{dataType}".""")
        series = series.drop_duplicates()
        series = series.sort_values()
        fpath2 = runOutputDir.joinpath("Set Files", f"{VARIABLE_NAME_TO_FILE_NAME_DICT[columnName]}.txt")
        makeDirPath(fpath2.parent)
        series.to_csv(fpath2, quoting=quoting, index=False, header=False)

    # Return path to sets fo ID values
    # TODO If this is implemented as a function, instead of a stand-alone script, return `runOutputDir` to define `setsPathDir` in the "makeMap" scripts.
    logging.info(f"""Finished collecting the set of ID values to de-identify. The set files are located in "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Collects i2b2 `PATIENT_NUM` and `ENCOUNTER_NUM` values.

NOTE This is not needed if you use the output from i2b2_dump.get_IDs().
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
I2B2_PORTION_OUTPUT_DIR_PATH = Path(r"..\Intermediate Results\i2b2 Portion\data\output\i2b2_dump\2023-11-20 20-35-58\i2b2")

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"
connection = create_engine(conStr).connect().execution_options(stream_results=True)

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    `I2B2_PORTION_OUTPUT_DIR_PATH`: "{I2B2_PORTION_OUTPUT_DIR_PATH}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Get set of all patient and encounter IDs
    logger.info("Getting the set of all patient and encounter IDs.")
    patientpath = runOutputDir.joinpath("Sets", "PATIENT_NUM.CSV")
    encounterpath = runOutputDir.joinpath("Sets", "ENCOUNTER_NUM.CSV")
    makeDirPath(patientpath.parent)
    makeDirPath(encounterpath.parent)
    patientmode = "w"
    encountermode = "w"
    patientheader = True
    encounterheader = True
    MESSAGE_MODULO = 50
    for fpath in I2B2_PORTION_OUTPUT_DIR_PATH.iterdir():
        logger.info(f"""  Working on file "{fpath.absolute().relative_to(rootDirectory)}".""")
        if fpath.suffix.lower() == ".csv":
            numChunks = 0
            CHUNKSIZE = 10000
            for _ in pd.read_csv(fpath, chunksize=CHUNKSIZE):
                numChunks += 1
            if numChunks < MESSAGE_MODULO:
                numChunksTenth = numChunks
            else:
                numChunksTenth = round(numChunks / MESSAGE_MODULO)
            for it, chunk in enumerate(pd.read_csv(fpath, chunksize=CHUNKSIZE), start=1):
                if it % numChunksTenth == 0:
                    logger.info(f"""    Working on chunk {it} of {numChunks}.""")
                chunk = pd.DataFrame(chunk)
                if "PATIENT_NUM" in chunk.columns:
                    patientIDs = chunk["PATIENT_NUM"].drop_duplicates().sort_values()
                    patientIDs.to_csv(patientpath, mode=patientmode, header=patientheader, index=False)
                    patientmode = "a"
                    patientheader = False
                if "ENCOUNTER_NUM" in chunk.columns:
                    encounterIDs = chunk["ENCOUNTER_NUM"].drop_duplicates().sort_values()
                    encounterIDs.to_csv(encounterpath, mode=encountermode, header=encounterheader, index=False)
                    encountermode = "a"
                    encounterheader = False
        else:
            pass

    # Save sorted, unique IDs, and drop negative encounter IDs
    # Note that in i2b2, negative encounter IDs are actually negative patient IDs.
    logger.info("Saving sorted, unique IDs - patients.")
    patientset = pd.read_csv(patientpath)["PATIENT_NUM"].drop_duplicates().dropna().sort_values()
    logger.info("Saving sorted, unique IDs - patients - done.")

    logger.info("Saving sorted, unique IDs - encounters.")
    encounterset = pd.read_csv(encounterpath)["ENCOUNTER_NUM"].drop_duplicates().dropna().sort_values()
    logger.info("Saving sorted, unique IDs - encounters - done.")
    mask = encounterset > 0
    encounterset = encounterset[mask]

    patientset.to_csv(patientpath, index=False)
    encounterset.to_csv(encounterpath, index=False)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
"""

from logging import Logger
from pathlib import Path
from typing_extensions import Union
# Third-party packages
import pandas as pd
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.drapi import (makeChunks,
                                    readDataFile,
                                    replace_sql_query)
from drapi.code.drapi.getData.getData_inner import getData_inner


def getData_outer(conStr: str,
                filterVariableColumnName: str,
                filterVariableChunkSize: int,
                filterVariableData: Union[pd.DataFrame, pd.Series],
                filterVariableFilePath: Union[Path, str],
                filterVariablePythonDataType: str,
                filterVariableSqlQueryTemplatePlaceholder: str,
                outputName: str,
                queryChunkSize: int,
                runOutputDir: Path,
                sqlFilePath: Union[Path, str],
                logger: Logger) -> pd.DataFrame:
    """
    Executes a SQL query with a filter.
    """
    # >>> Determine filter variable data sourcing type: Direct or indirect sourcing >>> NOTE This block is a little rough around the edges. TODO Needs testing.
    # Note -> Direct sourcing
    # Note -> Direct sourcing: Case 1: `filterVariableData` is defined and is a Series
    # Note -> Direct sourcing: Case 2: `filterVariableData` is defined and is a DataFrame and is of width 1
    # Note -> Direct sourcing: Case 3: `filterVariableData` is defined and is a DataFrame and `filterVariableColumnName` is defined
    # Note -> Indirect sourcing
    # Note -> Indirect sourcing: Case 1: `filterVariableFilePath` and `filterVariableColumnName` are defined

    # Determine filter variable data sourcing type: Direct or indirect sourcing: Direct sourcing
    sourcing = None
    if isinstance(filterVariableData, type(None)):
        sourcing = "direct - False"
        sourcingCase = None
    elif isinstance(filterVariableData, (pd.DataFrame)):
        # If-block 1
        if filterVariableData.shape[1] == 1:
            sourcing = "direct"
            sourcingCase = 2
        else:
            sourcingCase = None
        # If-block 2
        if isinstance(filterVariableColumnName,  (str, int)):
            sourcing = "direct"
            sourcingCase = 3
        else:
            sourcing = "direct - False"
            sourcingCase = None
    elif isinstance(filterVariableData, (pd.Series)):
        sourcing = "direct"
        sourcingCase = 1
    else:
        message = f"""The variable `filterVariableData` is of an unexpected data type: "{filterVariableData}"."""
        logger.critical(message)
        raise Exception(message)

    # Determine filter variable data sourcing type: Direct or indirect sourcing: Indirect sourcing
    if sourcing == "direct":
        pass
    else:
        if isinstance(filterVariableFilePath, type(None)) or isinstance(filterVariableColumnName, type(None)):
            sourcing = "indirect - False"
            sourcingCase = None
        elif isinstance(filterVariableFilePath, (Path, str)) and isinstance(filterVariableColumnName, (str, int)):
            sourcing = "indirect"
            sourcingCase = 1
        else:
            message = f"""One of the following variables is of an unexpected data type: `filterVariableData` -> "{type(filterVariableData)}"; `filterVariableColumnName` -> "{type(filterVariableColumnName)}"."""
            logger.critical(message)
            raise Exception(message)
    # <<< Determine filter variable data sourcing type: Direct or indirect sourcing <<<

    # >>> Select filter variable data >>>
    if sourcing == "direct":
        if sourcingCase == 1:
            filterVariableDataSeries = filterVariableData
        elif sourcingCase == 2:
            filterVariableDataSeries = filterVariableData.iloc[:,0]
        elif sourcingCase == 3:
            filterVariableDataSeries = filterVariableData[filterVariableColumnName]
        else:
            message = "An unexpected error occurred when determining the direct sourcing case for the filter variable data. Check the values for `filterVariableData` and `filterVariableColumnName`."
            logger.critical(message)
            raise Exception(message)
    elif sourcing == "indirect":
            dataFrame = readDataFile(fname=Path(filterVariableFilePath))
            filterVariableDataSeries = dataFrame[filterVariableColumnName]
            del dataFrame
    else:
        message = f"""An unexpected error occurred when determinig the filter variable data sourcing type. Check the values for `filterVariableColumnName`, `filterVariableData`, and `filterVariableColumnName`.\nDebugging values:\n`sourcing`: "{sourcing}"\n`sourcingCase`: "{sourcingCase}" """
        logger.critical(message)
        raise Exception(message)
    # <<< Select filter variable data <<<
        
    connection1 = create_engine(conStr).connect().execution_options(stream_results=True)
    connection2 = create_engine(conStr).connect().execution_options(stream_results=True)

    # Read query file
    with open(sqlFilePath, "r") as file:
        query0 = file.read()

    # Fill out query template
    logger.info("""  Filling out query template.""")

    if filterVariablePythonDataType == "int":
        filterVariableValues = filterVariableDataSeries.drop_duplicates().dropna().astype("int64").sort_values().to_list()
    elif filterVariablePythonDataType == "str":
        filterVariableValues = filterVariableDataSeries.drop_duplicates().dropna().sort_values().to_list()
    else:
        raise Exception(f"""Unexpected value for ``: {type(filterVariablePythonDataType)}.""")

    chunkGenerator1 = makeChunks(filterVariableValues, filterVariableChunkSize)
    chunkGenerator2 = makeChunks(filterVariableValues, filterVariableChunkSize)
    numChunks1 = sum([1 for _ in chunkGenerator1])
    padlen1 = len(str(numChunks1))
    for it1, dfChunk in enumerate(chunkGenerator2, start=1):
        itstring1 = str(it1).zfill(padlen1)
        logger.info(f"""    Working on filter chunk {itstring1} of {numChunks1} with `filterVariableChunkSize` "{filterVariableChunkSize:,}".""")

        # Fill query template: lists to strings
        if filterVariablePythonDataType == "int":
            filterVariableValuesAsString = ",".join([f"{el}" for el in dfChunk])
        elif filterVariablePythonDataType == "str":
            filterVariableValuesAsString = ",".join([f"'{el}'" for el in dfChunk])

        query = replace_sql_query(query=query0,
                                  old="""(( ADMIT_EVENT_Derived.NUM_GRAM_WGHT )/1000)/((( ADMIT_EVENT_Derived.NUM_CENTMTR_HGHT )/100)*(( ADMIT_EVENT_Derived.NUM_CENTMTR_HGHT )/100)) as "Admit BMI",""",
                                  new="""(( ADMIT_EVENT_Derived.NUM_GRAM_WGHT )/1000)/NULLIF(((( ADMIT_EVENT_Derived.NUM_CENTMTR_HGHT )/100)*(( ADMIT_EVENT_Derived.NUM_CENTMTR_HGHT )/100)), 0) as "Admit BMI",""")
        query = replace_sql_query(query=query,
                                  old=",(cast(wt.last_wt_oz as decimal(10,2))*0.0283495)/((cast(ht.last_ht_in as decimal(10,2))*0.0254)*(cast(ht.last_ht_in as decimal(10,2)))*0.0254) as bmi_manual_calc",
                                  new=",(cast(wt.last_wt_oz as decimal(10,2))*0.0283495)/NULLIF(((cast(ht.last_ht_in as decimal(10,2))*0.0254)*(cast(ht.last_ht_in as decimal(10,2)))*0.0254), 0) as bmi_manual_calc")

        query = replace_sql_query(query=query,
                                  old=filterVariableSqlQueryTemplatePlaceholder,
                                  new=filterVariableValuesAsString)

        getData_inner(conStr=conStr,
                      logger=logger,
                      outputName=outputName,
                      queryChunkSize=queryChunkSize,
                      runOutputDir=runOutputDir,
                      sqlQuery=query,
                      itstring1=itstring1,
                      numChunks1=numChunks1,
                      sqlFilePath=None)

    connection1.close()
    connection2.close()



"""
Script for modifying clinical text TSV files so they can be handled by the SDOH pipeline. This is a parallelized implementation.
"""

import concurrent
import concurrent.futures
import logging
from pathlib import Path
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.modifyTSV import wrapModify

# Arguments
CLINICAL_TEXT_DIR_1 = "../../../Data Request 2 - 2023-07-29/Intermediate Results/Notes Portion/data/output/freeText/2024-04-03 12-38-35/free_text/FlaDia_note"
CLINICAL_TEXT_DIR_2 = "../../../Data Request 2 - 2023-07-29/Intermediate Results/Notes Portion/data/output/freeText/2024-04-03 12-38-35/free_text/FlaDia_order_impression"
CLINICAL_TEXT_DIR_3 = "../../../Data Request 2 - 2023-07-29/Intermediate Results/Notes Portion/data/output/freeText/2024-04-03 12-38-35/free_text/FlaDia_order_narrative"
CLINICAL_TEXT_DIR_4 = "../../../Data Request 2 - 2023-07-29/Intermediate Results/Notes Portion/data/output/freeText/2024-04-03 12-38-35/free_text/FlaDia_order_result_comment"
ALL_CLINICAL_TEXT_DIRECTORIES = [CLINICAL_TEXT_DIR_1,
                                 CLINICAL_TEXT_DIR_2,
                                 CLINICAL_TEXT_DIR_3,
                                 CLINICAL_TEXT_DIR_4]


# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 3
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 4

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
# ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
#  "IRB_DIRECTORY",                 # noqa
#  "DATA_REQUEST_DIRECTORY",        # noqa
#  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
pass

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    ``: "{"..."}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"
    """)

    # Script
    # Initialize argument container for parallel wrapper
    allPaths0 = []
    allTypes0 = []
    allToDirectory = []
    allLogger = []
    for directory in ALL_CLINICAL_TEXT_DIRECTORIES:
        clinicalTextDirectory = Path(directory)
        for fpath in clinicalTextDirectory.iterdir():
            # Argument 1
            allPaths0.extend([fpath])
            # Argument 2
            if "note" in clinicalTextDirectory.name:
                allTypes0.extend(["note"])
            elif "order" in clinicalTextDirectory.name:
                allTypes0.extend(["order"])
            else:
                allTypes0.extend([None])
            # Argument 3
            allToDirectory.extend([runOutputDir])
            # Argument 4
            allLogger.extend([logger])


    df0 = pd.DataFrame()
    df0["All Paths"] = allPaths0
    df0["All Types"] = allTypes0
    select = ~df0.isna().any(axis=1)
    df = df0[select]
    logger.info(f"We dropped {sum(~select)} paths because of invalid clinical text types.")

    allPaths = df["All Paths"].to_list()
    allTypes = df["All Types"].to_list()
    message = f"Data for parallelization:\n{df.to_string()}"
    logger.info(message)

    with concurrent.futures.ProcessPoolExecutor() as pool:
        results = zip(*pool.map(wrapModify, allPaths,
                                            allTypes,
                                            allToDirectory,
                                            allLogger))

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Aliases variables that need to be de-identified.

INPUT:
Set Files

PROCESS:
Merge alias Set Files under the aliased (main) variable name

OUTPUT:
A new set file. This file would be the input to "makeMapsFromOthers.py"
"""

import csv
import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import VARIABLE_NAME_TO_FILE_NAME_DICT, FILE_NAME_TO_VARIABLE_NAME_DICT
# Super-local
from common import VARIABLE_ALIASES, DATA_TYPES_DICT

# Arguments
SET_FILES_DIR = Path(r"..\Concatenated Results\data\output\getIDValues\...\Set Files")

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                    # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                    #  "IRB_DIRECTORY",                 # noqa
                                    #  "DATA_REQUEST_DIRECTORY",        # noqa
                                    #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"
connection = create_engine(conStr).connect().execution_options(stream_results=True)

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    `SET_FILES_DIR`: "{SET_FILES_DIR}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Script
    for setFilePath in SET_FILES_DIR.iterdir():
        logger.info(f"""  Working on set file : "{setFilePath.absolute().relative_to(rootDirectory)}".""")
        try:
            df = pd.read_table(setFilePath, header=None)
        except EmptyDataError as err:
            _ = err
            df = pd.DataFrame(dtype=str)
        variableName = FILE_NAME_TO_VARIABLE_NAME_DICT[setFilePath.stem]
        condition1 = variableName in VARIABLE_ALIASES.keys()
        condition2 = variableName in VARIABLE_ALIASES.values()
        dataType = DATA_TYPES_DICT[variableName]
        if dataType == "Datetime":
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "Numeric":
            df = df.astype(float).astype("Int64")
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "String":
            quoting = csv.QUOTE_ALL
        elif dataType == "Numeric_Or_String":
            quoting = csv.QUOTE_ALL
        else:
            raise Exception(f"""Unexpected `dataType` value: "{dataType}".""")
        if condition1 or condition2:
            if condition1:
                aliasName = variableName
                mainName = VARIABLE_ALIASES[aliasName]
                logger.info(f"""    Variable is aliased to "{mainName}".""")
            elif condition2:
                mainName = variableName
                logger.info("""    Variable is the main name of several aliases.""")
            setFileName = f"{mainName}.txt"
            newSetFilePath = runOutputDir.joinpath(setFileName)
            if newSetFilePath.exists():
                logger.info("""  ..  Variable set exists. Appending to file.""")
                fileMode = "a"
            else:
                logger.info("""  ..  Variable set does not exists. Creating file.""")
                fileMode = "w"
            df.to_csv(newSetFilePath, quoting=quoting, mode=fileMode, index=False, header=False)
        else:
            logger.info("""    Variable set file is not aliased.""")
            setFileName = f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]}.txt"
            newSetFilePath = runOutputDir.joinpath(setFileName)
            df.to_csv(newSetFilePath, quoting=quoting, index=False, header=False)

    # Sort and remove duplicates from merged results.
    logger.info("Sorting data and removing duplicates.")
    listOfFiles = sorted(runOutputDir.iterdir())
    for setFilePath in listOfFiles:
        logger.info(f"""  Working on variable `setFilePath`: "{setFilePath.absolute().relative_to(rootDirectory)}".""")
        try:
            series = pd.read_table(setFilePath, header=None).iloc[:, 0]
        except EmptyDataError as err:
            _ = err
            series = pd.Series(dtype=str)
        logger.info(f"""    Starting size of variable set:               "{len(series):,}".""")
        variableName = FILE_NAME_TO_VARIABLE_NAME_DICT[setFilePath.stem]
        dataType = DATA_TYPES_DICT[variableName]
        if dataType == "Datetime":
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "Numeric":
            series = series.astype(float).astype("Int64")
            quoting = csv.QUOTE_MINIMAL
        elif dataType == "String":
            quoting = csv.QUOTE_ALL
        elif dataType == "Numeric_Or_String":
            quoting = csv.QUOTE_ALL
        else:
            raise Exception(f"""Unexpected `dataType` value: "{dataType}".""")
        series = series.drop_duplicates().sort_values()
        logger.info(f"""    Variable set size after removing duplicates: "{len(series):,}".""")
        series.to_csv(setFilePath, quoting=quoting, index=False, header=False)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Makes de-identification maps, building on existing maps.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects integer IDs, so no string IDs like Epic Patient IDs.
"""

import json
import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.code.drapi.drapi import (ditchFloat,
                                    getTimestamp,
                                    handleDatetimeForJson,
                                    makeDirPath,
                                    makeMap,
                                    makeSetComplement,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import (FILE_NAME_TO_VARIABLE_NAME_DICT,
                                          VARIABLE_NAME_TO_FILE_NAME_DICT)
# Local packages: Script parameters: General
from common import (IRB_NUMBER,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    DATA_TYPES_DICT,
                    OLD_MAPS_DIR_PATH,
                    VARIABLE_ALIASES,
                    VARIABLE_SUFFIXES)

# Arguments
SETS_PATH = Path(r"..\Concatenated Results\data\output\aliasVariables\...")

CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in SETS_PATH.iterdir()])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        getIDValuesOutput = SETS_PATH
    elif operatingSystem == "win32":
        getIDValuesOutput = SETS_PATH
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Get set of values
    # NOTE The code that used to be in this section was moved to "getIDValues.py"
    logging.info(f"""Using the set of new values in the directory "{getIDValuesOutput.absolute().relative_to(rootDirectory)}".""")

    # QA: Make sure all data variables are present in the script parameters.
    collectedVariables = [FILE_NAME_TO_VARIABLE_NAME_DICT[fname.stem] for fname in getIDValuesOutput.iterdir()]
    missingDataTypes = []
    missingVariableSuffixes = []
    for variableName in collectedVariables:
        if variableName not in DATA_TYPES_DICT.keys():
            missingDataTypes.append(variableName)
        if variableName not in VARIABLE_SUFFIXES.keys():
            missingVariableSuffixes.append(variableName)
    if len(missingDataTypes) > 0:
        text = "\n".join(missingDataTypes)
        raise Exception(f"""Not all variables have a data type assigned to them:\n{text}""")
    if len(missingVariableSuffixes) > 0:
        text = "\n".join(missingVariableSuffixes)
        raise Exception(f"Not all variables have a de-identification suffix assigned to them:\n{text}")

    # Create reverse-look-up alias map
    variableAliasesReverse = {}
    for variableAlias, variableMainName in VARIABLE_ALIASES.items():
        if variableMainName in variableAliasesReverse.keys():
            variableAliasesReverse[variableMainName].append(variableAlias)
        else:
            variableAliasesReverse[variableMainName] = [variableAlias]

    # Concatenate all old maps
    oldMaps = {}
    logging.info("""Concatenating all similar pre-existing maps.""")
    for variableName in collectedVariables:
        logging.info(f"""  Working on variable "{variableName}".""")
        # Get maps explicitely or implicietly referring to this variable name.
        condition1 = variableName in OLD_MAPS_DIR_PATH.keys()
        if condition1:
            variableLookupName = variableName
            logging.info("""    Variable has pre-existing map(s).""")
            listOfMapPaths = OLD_MAPS_DIR_PATH[variableLookupName]
            dfConcat = pd.DataFrame()
            for mapPath in listOfMapPaths:
                logging.info(f"""  ..  Reading pre-existing map from "{mapPath.absolute().relative_to(rootDirectory)}".""")
                df = pd.DataFrame(pd.read_csv(mapPath))
                dfConcat = pd.concat([dfConcat, df])
            oldMaps[variableName] = dfConcat
        if variableName in variableAliasesReverse.keys():
            for variableAlias in variableAliasesReverse[variableName]:
                condition2 = variableAlias in OLD_MAPS_DIR_PATH.keys()
                if condition2:
                    variableLookupName = variableAlias
                    logging.info("""    Variable has pre-existing aliased map(s).""")
                    listOfMapPaths = OLD_MAPS_DIR_PATH[variableLookupName]
                    dfConcat = pd.DataFrame()
                    for mapPath in listOfMapPaths:
                        logging.info(f"""  ..  Reading pre-existing map from "{mapPath}".""")
                        df = pd.DataFrame(pd.read_csv(mapPath))
                        dftemp = makeMap(IDset=set(), IDName=variableName, startFrom=0, irbNumber=IRB_NUMBER, suffix="", columnSuffix=variableName, deIdentificationMapStyle="lemur", logger=logging.getLogger())
                        df.columns = dftemp.columns
                        dfConcat = pd.concat([dfConcat, df])
                    if condition1:
                        df = oldMaps[variableName]
                        dfConcat = pd.concat([dfConcat, df])
                        oldMaps[variableName] = dfConcat
                    else:
                        oldMaps[variableName] = dfConcat
        else:
            condition2 = False
        if not (condition1 or condition2):
            logging.info("""    Variable has no pre-existing map.""")
            oldMaps[variableName] = pd.DataFrame()

    # Get the set difference between all old maps and the set of un-mapped values
    valuesToMap = {}
    setsToMapDataDir = runIntermediateDataDir.joinpath("valuesToMap")
    makeDirPath(setsToMapDataDir)
    logging.info("""Getting the set difference between all old maps and the set of un-mapped values.""")
    for variableName in collectedVariables:
        logging.info(f"""  Working on variable "{variableName}".""")
        variableDataType = DATA_TYPES_DICT[variableName]

        # Get old set of IDs
        logging.info("""    Getting the old set of IDs.""")
        oldMap = oldMaps[variableName]
        oldMap = pd.DataFrame(oldMap)
        if len(oldMap) > 0:
            if variableDataType.lower() == "numeric":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.unique()
                oldIDSet = set([ditchFloat(el) for el in oldIDSet])  # NOTE: Hack. Convert values to type int or string
            elif variableDataType.lower() == "string":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.astype(str)
                oldIDSet = oldIDSet.unique()
            elif variableDataType.lower() == "numeric_or_string":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.astype(str)
                oldIDSet = oldIDSet.unique()
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)
        elif len(oldMap) == 0:
            oldIDSet = set()
        logging.info(f"""    The size of this set is {len(oldIDSet):,}.""")

        # Get new set of IDs
        newSetPath = getIDValuesOutput.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]}.txt")
        logging.info(f"""    Getting the new set of IDs from "{newSetPath.absolute().relative_to(rootDirectory)}".""")
        try:
            setSeries = pd.read_table(newSetPath, header=None)[0]
            newIDSet = set(setSeries.to_list())
        except EmptyDataError as err:
            _ = err
            newIDSet = set()
        if variableDataType.lower() == "numeric":
            newIDSet = set([ditchFloat(el) for el in newIDSet])  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            newIDSet = set([str(el) for el in newIDSet])
        elif variableDataType.lower() == "numeric_or_string":
            newIDSet = set([str(el) for el in newIDSet])
        else:
            msg = "The table column is expected to have a data type associated with it."
            logging.error(msg)
            raise Exception(msg)
        logging.info(f"""    The size of this set is     {len(newIDSet):,}.""")

        # Set difference
        IDSetDiff = newIDSet.difference(oldIDSet)
        logging.info(f"""    The set difference size is  {len(IDSetDiff):,}.""")
        valuesToMap[variableName] = IDSetDiff

        # Save new subset to `setsToMapDataDir`
        fpath = setsToMapDataDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]}.JSON")
        with open(fpath, "w") as file:
            if variableDataType.lower() == "numeric":
                li = [ditchFloat(IDNumber) for IDNumber in IDSetDiff]  # NOTE: Hack. Convert values to type int or string
            elif variableDataType.lower() == "string":
                li = list(IDSetDiff)
            elif variableDataType.lower() == "numeric_or_string":
                li = list(IDSetDiff)
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)
            file.write(json.dumps(li, default=handleDatetimeForJson))
        if len(IDSetDiff) == 0:
            series = pd.Series(dtype=int)
        else:
            if variableDataType.lower() == "numeric":
                series = pd.Series(sorted(list(IDSetDiff)))
            elif variableDataType.lower() == "string":
                series = pd.Series(sorted([str(el) for el in IDSetDiff]))
            elif variableDataType.lower() == "numeric_or_string":
                series = pd.Series(sorted([str(el) for el in IDSetDiff]))
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)

    # Get numbers for new map
    logging.info("""Getting numbers for new map.""")
    newNumbersDict = {}
    for variableName in collectedVariables:
        oldMap = oldMaps[variableName]
        if len(oldMap) > 0:
            oldNumbersSet = set(oldMap.iloc[:, 1].values)
        elif len(oldMap) == 0:
            oldNumbersSet = set()
        # Get quantity of numbers needed for map
        quantityOfNumbersUnmapped = len(valuesToMap[variableName])
        # Get new numbers
        lenOlderNumbersSet = len(oldNumbersSet)
        if lenOlderNumbersSet == 0:
            newNumbers = list(range(1, quantityOfNumbersUnmapped + 1))
        else:
            newNumbersSet = makeSetComplement(oldNumbersSet, quantityOfNumbersUnmapped)
            newNumbers = sorted(list(newNumbersSet))
        newNumbersDict[variableName] = newNumbers

    # Map un-mapped values
    logging.info("""Mapping un-mapped values.""")
    for file in setsToMapDataDir.iterdir():
        variableName = FILE_NAME_TO_VARIABLE_NAME_DICT[file.stem]
        variableDataType = DATA_TYPES_DICT[variableName]
        logging.info(f"""  Working on un-mapped values for variable "{variableName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Map contents
        values = valuesToMap[variableName]
        if variableDataType.lower() == "numeric":
            values = set(int(float(value)) for value in values)  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            pass
        elif variableDataType.lower() == "numeric_or_string":
            pass
        else:
            msg = "The table column is expected to have a data type associated with it."
            logging.error(msg)
            raise Exception(msg)
        numbers = sorted(list(newNumbersDict[variableName]))
        deIdIDSuffix = VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"]
        map_ = makeMap(IDset=values,
                       IDName=variableName,
                       startFrom=numbers,
                       irbNumber=IRB_NUMBER,
                       suffix=deIdIDSuffix,
                       columnSuffix=variableName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        # Save map
        mapPath = runOutputDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]} map.csv")
        map_.to_csv(mapPath, index=False)
        logging.info(f"""    De-identification map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Outline

    - Initilalize an empty pandas dataframe as the master de-identification map.
    - Iterate over the encounter SQLite database files.
    - For each file, append the encounter de-identification maps to the master map.
"""

""" NOTE See DNR folder for latest version
# Private variables ("dunders")
__all__ = ["encounterMapDf"]

# Imports
from pathlib import Path
import logging
import os
import re
import sys
# Third-party packages
import numpy as np
import pandas as pd
import sqlite3
# Local imports
from drapi.drapi import sqlite2df, getTimestamp, replace_sql_query, makeChunks

# Arguments
QUERY_PATH = Path("sql/encounterNumber2patientKey.SQL")
DATABSE_DIR_MAC = os.path.join("/",
                               "Volumes",
                               "FILES",
                               "FTP",
                               "IDR",
                               "ANES",
                               "IRB201600223 - aka R01",
                               "Deiden_db")
DATABSE_DIR_WINDOWS = os.path.join("X:\\",
                                   "FTP",
                                   "IDR",
                                   "ANES",
                                   "IRB201600223 - aka R01",
                                   "Deiden_db")
CHUNK_SIZE = "100,000"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir = thisFilePath.absolute().parent.parent
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: more
chunkSize = int(CHUNK_SIZE.replace(",", ""))

# Connect to SQLite database
if True:
    # If you have connection to the below directory, use the below line.
    operatingSystem = sys.platform
    if operatingSystem == "win32":
        database_dir = Path(DATABSE_DIR_WINDOWS)
    elif operatingSystem == "darwin":
        database_dir = Path(DATABSE_DIR_MAC)
    else:
        raise Exception("Unsupported operating system")
elif True:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    database_dir = None


def getDeidenDBSeriesNumber(string):
    """
    Regular expression to get the series number from the IDEALIST encounter de-identification database file.

    Expects files to be of the format "deiden_YYYY-MM-DD_XX.db", where YYYY is the year four-digit year, MM is the two-digit month, DD is the two-digit day, and XX is the series number (with no zero padding).
    """
    seriesNumber = re.search(r"_(\d+).db$", string).groups(0)[0]
    return seriesNumber


if __name__ == "__main__":
    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `projectDir`: "{projectDir}".""")

    # SQLite connection
    databaseFiles = sorted([fPath for fPath in database_dir.iterdir() if "deiden_2021-05-01_" in fPath.name], key=lambda fPath: int(getDeidenDBSeriesNumber(str(fPath))))
    encounterMapDf = pd.DataFrame()
    for it, file in enumerate(databaseFiles, start=1):
        database_path = file
        logging.info(f"""  Loading file {it} of {len(databaseFiles)}.""")
        logging.info(f"""    Loading sqlite database from "{database_path}".""")
        sqliteConnection = sqlite3.connect(database_path)
        cursor = sqliteConnection.cursor()

        if it == 1:
            # Test SQLite connection
            logging.info("""  ..  Testing SQLite connection.""")
            query = """SELECT 1"""
            cursor.execute(query)
            test1 = cursor.fetchall()[0][0]
            if test1:
                logging.info(f"""  ..    SQLite connection successful: "{test1}".""")

        # Query encounter de-identification map
        logging.info("""    Running SQLite query for encounter map.""")
        query = """SELECT *
    FROM EncounterDeidenMap"""
        cursor.execute(query)
        resultsList = cursor.fetchall()
        logging.info("""    SQLite query completed.""")
        results = sqlite2df(resultsList, "EncounterDeidenMap", cursor)
        encounterMapDf = pd.concat([encounterMapDf, results])

    # Post-processing
    logging.info("""Ordering encounter de-identification map dataframe by values.""")
    encounterMapDf = encounterMapDf.sort_values(by="real_id")

    # Map encounter numbers to patient keys, in parallel / series
    logging.info("""Reading SQL file.""")
    with open(QUERY_PATH, "r") as file:
        query0 = file.read()

    indexChunks = makeChunks(array_range=range(len(encounterMapDf)), chunkSize=chunkSize)
    numChunks = int(np.ceil(len(encounterMapDf) / chunkSize))
    allResults = pd.DataFrame()
    for it, indices in enumerate(indexChunks, start=1):
        logging.info(f"""  Working on chunk {it} of {numChunks}.""")
        # Make chunk
        logging.info("""    Making chunk.""")
        indices = list(indices)
        dfChunk = encounterMapDf.iloc[indices, :]
        # Make query string from chunk
        logging.info("""    Making query string.""")
        series = dfChunk["real_id"]
        # li = sorted(series.to_list())
        li = series.to_list()
        encountersListAsString = ",".join([str(int(el)) for el in li])
        query = replace_sql_query(query0, "0123456789, 1234567890", encountersListAsString)
        # Query chunk
        logging.info("""    Querying database.""")
        results = pd.read_sql(query, con=conStr)
        # results.sort_values(by="ENCNTR_CSN_ID", key=int)
        allResults = pd.concat([allResults, results])

    # Create final de-identification map.
    if False:
        COLUMN_NAME = None
        encounterMapDf[COLUMN_NAME] = encounterMapDf["deiden_id"].apply(lambda integer: f"{integer}")

    # Script end
    logging.info(f"""Finished running "{thisFilePath}".""")
"""


"""
QA: Make sure values were correctly converted.
"""

import argparse
import logging
import os
import pprint
from pathlib import Path
from typing_extensions import Union
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    successiveParents)
# Super-local
pass


# Functions

def spotCheckColumns(listOfFiles1: list,
                     listofFiles2: list,
                     logger: logging.Logger,
                     moduloChunks: int,
                     columnsToCheck1: dict[int: str] = {0: "visit_occurrence_id",
                                                        1: "person_id",
                                                        2: "provider_id"},
                     columnsToCheck2: dict[int: str] = {0: "Encounter # (CSN)",
                                                        1: "Patient Key",
                                                        2: "Provider Key"}) -> None:
    """
    This function "spot checks" pairs of files to see if the passed variables have been properly converted from the OMOP data model to the UF Health data model.

    In `columnsToCheck1` and `columnsToCheck2` the keys of the dictionary have the following meaning:
    - 0: Encounter ID Variable Name
    - 1: Patient ID Variable Name
    - 2: Provider ID Variable Name

    The code is only meant to check these three ID types, because under the hood these integer values are mapped to SQL queries. So, for example, you could not pass `{0: "Note Key"}` to the function, because it would use the SQL query that maps encounter IDs.

    NOTE This has only been tested and verified on 3-on-3 checks. That is, checking if all three informative OMOP ID variables have been mapped to their three UF Health equivalents. The code should be able to also do spot checks for less number of variables, like just `person_id` to `Patient Key`, but it has not been tested.
    """
    with open("../Concatenated Results/sql/BO Variable Spot Check - Encounter # (CSN).sql", "r") as file:
        query0_0 = file.read()
    with open("../Concatenated Results/sql/BO Variable Spot Check - Patient Key.sql", "r") as file:
        query1_0 = file.read()
    with open("../Concatenated Results/sql/BO Variable Spot Check - Provider Key.sql", "r") as file:
        query2_0 = file.read()

    for file1, file2 in zip(listOfFiles1, listofFiles2):
        try:
            logger.info(f"""  Working on file pair:
        `file1`: "{file1}"
        `file2`: "{file2}".""")
            df2_0 = pd.read_csv(file2, nrows=2)
            df2_0columns = df2_0.columns
            columnChecks2 = df2_0columns.isin(columnsToCheck2.values())
            logger.debug(f"""  Group 2 table:\n{df2_0.to_string()}""")
            logger.debug(f"""  Group 2 columns:\n{df2_0.columns}""")
            logger.debug(f"""  Group 2 columns check:\n{columnChecks2}""")
            spot2df = df2_0.loc[0, columnChecks2]
            spot2df = pd.DataFrame(spot2df)  # For type hinting
            logger.info(f"  A row has been selected for a spot check:\n{spot2df.T.to_string()}")
            # Convert group 2 spot to group 1 values
            spot2dict = spot2df.to_dict()[0]
            logger.info(spot2dict)
            query0_1 = query0_0[:]
            query1_1 = query1_0[:]
            query2_1 = query2_0[:]
            spot1dict = {}
            if 0 in columnsToCheck2.keys():
                encounterIDVariableName = columnsToCheck2[0]
                if df2_0columns.isin([encounterIDVariableName]).sum() > 0:
                    queryEncounterIDValue = spot2dict[encounterIDVariableName]
                    query0_1 = query0_1.replace("{PYTHON_VARIABLE: Encounter # (CSN)}", str(queryEncounterIDValue))
                    result0 = pd.read_sql(sql=query0_1, con=conStr)
                    if result0.shape != (1,2):
                        logger.warning(f"""  ..  The encounter ID mapping is not one-to-one:\n{result0.to_string()}.""")
                        spot1EncounterID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1EncounterID = result0["visit_occurrence_id"][0]
                    spot1dict["visit_occurrence_id"] = [spot1EncounterID]
            if 1 in columnsToCheck2.keys():
                patientIDVariableName = columnsToCheck2[1]
                if df2_0columns.isin([patientIDVariableName]).sum() > 0:
                    queryPatientIDValue = spot2dict[patientIDVariableName]
                    query1_1 = query1_1.replace("{PYTHON_VARIABLE: Patient Key}", str(queryPatientIDValue))
                    result1 = pd.read_sql(sql=query1_1, con=conStr)
                    if result1.shape != (1,2):
                        logger.warning(f"""  ..  The patient ID mapping is not one-to-one:\n{result1.to_string()}.""")
                        spot1PatientID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1PatientID = result1["person_id"][0]
                    spot1dict["person_id"] = [spot1PatientID]
            if 1 in columnsToCheck2.keys():
                providerIDVariableName = columnsToCheck2[2]
                if df2_0columns.isin([providerIDVariableName]).sum() > 0:
                    queryProviderIDValue = spot2dict[providerIDVariableName]
                    query2_1 = query2_1.replace("{PYTHON_VARIABLE: Provider Key}", str(queryProviderIDValue))
                    result2 = pd.read_sql(sql=query2_1, con=conStr)
                    if result2.shape != (1,2):
                        logger.warning(f"""  ..  The provider ID mapping is not one-to-one:\n{result2.to_string()}.""")
                        spot1ProviderID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1ProviderID = result2["provider_id"][0]
                    spot1dict["provider_id"] = [spot1ProviderID]
            spot1Targetdf = pd.DataFrame.from_dict(data=spot1dict, orient="columns")
            logger.info(f"""The values that we are looking for in the group 1 set of files is below:\n{spot1Targetdf}""" )
            # Chunk
            chunkGenerator1 = pd.read_csv(file1, chunksize=CHUNKSIZE, low_memory=False)
            chunkGenerator2 = pd.read_csv(file1, chunksize=CHUNKSIZE, low_memory=False)
            logger.info("    Counting the number of chunks in the file.")
            it1Total = sum([1 for _ in chunkGenerator1])
            logger.info(f"    Counting the number of chunks in the file - done. There are {it1Total:,} chunks.")
            if it1Total < MESSAGE_MODULO_CHUNKS:
                moduloChunks = it1Total
            else:
                moduloChunks = round(it1Total / MESSAGE_MODULO_CHUNKS)
            if it1Total / moduloChunks < 100:
                moduloChunks = 1
            else:
                pass
            for it1, df1Chunk in enumerate(chunkGenerator2, start=1):
                if it1 % moduloChunks == 0:
                    allowLogging = True
                else:
                    allowLogging = False
                if allowLogging:
                    logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
                df1Chunk = pd.DataFrame(df1Chunk)  # For type hinting
                columnChecks1 = df1Chunk.columns.isin(columnsToCheck1.values())
                df1 = df1Chunk.loc[:, columnChecks1]
                df1 = pd.DataFrame(df1)  # For type hinting
                mask = df1.isin(spot1Targetdf.values.flatten()).all(axis=1)
                rowLocation = mask.index[mask][0]
                if mask.sum() > 0:
                    spot1Hitdf = df1[mask]
                    logger.info(f"""    Found spot located at index value (row number) "{rowLocation}":\n{spot1Hitdf.to_string()}.""")
                    break
        except Exception as error:
            logger.fatal(error)


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--Encounter_ID_1",
                        default="visit_occurrence_id",
                        type=str,
                        choices=["visit_occurrence_id"],
                        help="")
    parser.add_argument("--Encounter_ID_2",
                        default="Encounter # (CSN)",
                        type=str,
                        choices=["Encounter # (CSN)"],
                        help="")
    parser.add_argument("--Patient_ID_1",
                        default="person_id",
                        type=str,
                        choices=["person_id"],
                        help="")
    parser.add_argument("--Patient_ID_2",
                        default="Patient Key",
                        type=str,
                        choices=["Patient Key",
                                 "PatientKey"],
                        help="")
    parser.add_argument("--Provider_ID_1",
                        default="provider_id",
                        type=str,
                        choices=["provider_id"],
                        help="")
    parser.add_argument("--Provider_ID_2",
                        default="Provider Key",
                        type=str,
                        choices=["Provider Key"],
                        help="")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-variables
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    encounterID1 = argNamespace.Encounter_ID_1
    patientID1 = argNamespace.Patient_ID_1
    providerID1 = argNamespace.Provider_ID_1
    encounterID2 = argNamespace.Encounter_ID_2
    patientID2 = argNamespace.Patient_ID_2
    providerID2 = argNamespace.Provider_ID_2

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-variables
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `CHUNKSIZE`: "{CHUNKSIZE}"
    `MESSAGE_MODULO_CHUNKS`: "{MESSAGE_MODULO_CHUNKS}"
    `MESSAGE_MODULO_FILES`: "{MESSAGE_MODULO_FILES}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USER_DOMAIN` = "{USER_DOMAIN}"
    `USERNAME` = "{USERNAME}"
    `USER_ID` = "{USER_ID}"
    `USER_PWD` = censored
    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    TEST_SCRIPT = False

    if not TEST_SCRIPT:
        LIST_1 = ["/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Intermediate Results/OMOP Portion/data/output/identified/2024-01-23 13-54-26/condition_occurrence.csv"] + [path.absolute() for path in Path("/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/OMOP_Structured_Data/data/output/identified").iterdir() if path.suffix == ".csv"]
        LIST_2 = [path.absolute() for path in Path("../Concatenated Results/data/output/convertColumnsGeneral/2024-02-23 18-11-43").iterdir()]
    elif TEST_SCRIPT:
        LIST_1 = ["/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/OMOP_Structured_Data/data/output/identified/device_exposure.csv"]
        LIST_2 = ["/data/herman/Projects/IRB202300703/Concatenated Results/data/output/convertColumnsGeneral/2024-02-23 18-11-43/device_exposure.csv"]

    listOfFiles1 = sorted([Path(string) for string in LIST_1])
    listOfFiles2 = sorted([Path(string) for string in LIST_2])

    lof1 = [pathObj.name for pathObj in listOfFiles1]
    lof2 = [pathObj.name for pathObj in listOfFiles2]

    dfFiles = pd.DataFrame([lof1, lof2]).T
    dfFiles.columns = ["Group 1 Files", "Group 2 Files"]

    logger.info(f"\n{dfFiles}")

    columnsToCheck1 = {integer: string for integer, string in [(0, encounterID1), (1, patientID1), (2, providerID1)] if not isinstance(string, type(None))}
    columnsToCheck2 = {integer: string for integer, string in [(0, encounterID2), (1, patientID2), (2, providerID2)] if not isinstance(string, type(None))}

    spotCheckColumns(listOfFiles1=listOfFiles1,
                     listofFiles2=listOfFiles2,
                     columnsToCheck1=columnsToCheck1,
                     columnsToCheck2=columnsToCheck2,
                     logger=logger,
                     moduloChunks=MESSAGE_MODULO_CHUNKS)

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Makes de-identification maps from scratch.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
# TODO Needs to combine similar IDs, like different providers IDs.
"""

from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.drapi import makeMap, getTimestamp, makeDirPath


def makeMapsFromScratch(SETS_PATH,
                        VARIABLE_SUFFIXES,
                        IRB_NUMBER,
                        pipelineOutputDir: Path,
                        logger,
                        ROOT_DIRECTORY,
                        rootDirectory):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Map values
    for file in SETS_PATH.iterdir():
        variableName = file.stem
        logger.info(f"""  Working on variable "{variableName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Read file
        try:
            df = pd.read_table(file, header=None)
        except EmptyDataError as err:
            _ = err
            df = pd.DataFrame()
        # Assert
        if df.shape[1] == 1:
            # Try to convert to integer-type
            try:
                df.iloc[:, 0] = df.iloc[:, 0].astype(int)
            except ValueError as err:
                _ = err
            # Check length differences
            len0 = len(df)
            values = set(df.iloc[:, 0].values)
            len1 = len(values)
            logger.info(f"""    The length of the ID array was reduced from {len0:,} to {len1:,} when removing duplicates.""")
        elif df.shape[1] == 0:
            pass
        # Map contents
        deIdIDSuffix = VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"]
        map_ = makeMap(IDset=values, IDName=variableName, startFrom=1, irbNumber=IRB_NUMBER, suffix=deIdIDSuffix, columnSuffix=variableName)
        # Save map
        mapPath = runOutputDir.joinpath(f"{variableName} map.csv")
        map_.to_csv(mapPath, index=False)
        logger.info(f"""    De-identification map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")
